/*
    Ruler.cs - ruler for NRuler
    Copyright (C) 2007, 2008 siliconforks.com

    This file is part of NRuler.

    NRuler is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

using System;
using System.Drawing;
using System.Windows.Forms;

namespace SiliconForks.NRuler {
  public class Ruler: Form {
    private Options options = new Options();
    private OptionsForm optionsForm = new OptionsForm();
    private Form help = new Help();

    private bool isOpaque = false;

    private int length = 600;

    private int breadth = 80;

    private Orientation orientation = Orientation.Horizontal;

    private bool isMoving = false;
    private bool isResizing = false;

    private Point initialDesktopLocation;
    private Point initialMouseDesktopLocation;
    private int previousX = -1;
    private int previousY = -1;
    private int previousLength = -1;

    private Font privateFont;

    private MenuItem horizontalMenuItem;
    private MenuItem verticalMenuItem;
    private MenuItem opaqueMenuItem;
    private MenuItem alwaysOnTopMenuItem;

    public Ruler() {
      Text = "NRuler";
      Icon = new Icon(typeof(Ruler), "Icon");

      optionsForm.Options = options;

      // Under Mono, minimum window width is 110
      if (Type.GetType("Mono.Runtime") != null) {
        breadth = 110;
      }

      // Under Mono, changing the default size causes the entire form to resize???!?
      // Font = new Font(FontFamily.GenericSansSerif, 9, GraphicsUnit.Pixel);
      privateFont = new Font(FontFamily.GenericSansSerif, 9, GraphicsUnit.Pixel);

      // double buffering
      SetStyle(ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint | ControlStyles.DoubleBuffer, true);
      
      // no title bar
      FormBorderStyle = FormBorderStyle.None;

      // set the size
      resize();

      // transparent
      Opacity = options.Opacity;

      // right-click menu
      ContextMenu = new ContextMenu();
      MenuItem helpMenuItem = new MenuItem("H&elp");
  //    helpMenuItem.Shortcut = Shortcut.F1;
      helpMenuItem.Click += new EventHandler(helpMenuItem_Click);
      ContextMenu.MenuItems.Add(helpMenuItem);
      horizontalMenuItem = new MenuItem("&Horizontal");
  //    horizontalMenuItem.Shortcut = Shortcut.CtrlH;
      horizontalMenuItem.Click += new EventHandler(horizontalMenuItem_Click);
      horizontalMenuItem.Enabled = false;
      ContextMenu.MenuItems.Add(horizontalMenuItem);
      verticalMenuItem = new MenuItem("&Vertical");
  //    verticalMenuItem.Shortcut = Shortcut.CtrlV;
      verticalMenuItem.Click += new EventHandler(verticalMenuItem_Click);
      ContextMenu.MenuItems.Add(verticalMenuItem);
      opaqueMenuItem = new MenuItem("&Opaque");
      opaqueMenuItem.Click += new EventHandler(opaqueMenuItem_Click);
      ContextMenu.MenuItems.Add(opaqueMenuItem);
      alwaysOnTopMenuItem = new MenuItem("Always on &Top");
  //    alwaysOnTopMenuItem.Shortcut = Shortcut.CtrlT;
      alwaysOnTopMenuItem.Click += new EventHandler(alwaysOnTopMenuItem_Click);
      ContextMenu.MenuItems.Add(alwaysOnTopMenuItem);

      MenuItem optionsMenuItem = new MenuItem("O&ptions");
      optionsMenuItem.Click += new EventHandler(optionsMenuItem_Click);
      ContextMenu.MenuItems.Add(optionsMenuItem);

      MenuItem exitMenuItem = new MenuItem("E&xit");
  //    exitMenuItem.Shortcut = Shortcut.CtrlQ;
      exitMenuItem.Click += new EventHandler(exitMenuItem_Click);
      ContextMenu.MenuItems.Add(exitMenuItem);
    }

    private void resize() {
      Invalidate();  // mono needs this
      if (length < 200) {
        length = 200;
      }
      if (orientation == Orientation.Horizontal) {
        ClientSize = new Size(length, breadth);
      }
      else {
        ClientSize = new Size(breadth, length);
      }
      Text = "NRuler (" + length + " px)";
    }

    private bool isOverHandle(MouseEventArgs e) {
      int l;
      if (orientation == Orientation.Horizontal) {
        l = e.X;
      }
      else {
        l = e.Y;
      }
      return l >= length - 10;
    }

    private void toggleHelp() {
      help.Show();
    }

    private void horizontal() {
      orientation = Orientation.Horizontal;
      horizontalMenuItem.Enabled = false;
      verticalMenuItem.Enabled = true;
      resize();
    }

    private void vertical() {
      orientation = Orientation.Vertical;
      horizontalMenuItem.Enabled = true;
      verticalMenuItem.Enabled = false;
      resize();
    }

    private void toggleOpaque() {
      if (isOpaque) {
        isOpaque = false;
        Opacity = options.Opacity;
        opaqueMenuItem.Checked = false;
      }
      else {
        isOpaque = true;
        Opacity = 1.0;
        opaqueMenuItem.Checked = true;
      }
    }

    private void toggleAlwaysOnTop() {
      TopMost = ! TopMost;
      alwaysOnTopMenuItem.Checked = TopMost;
    }

    private void quit() {
      options.Write();
      Application.Exit();
    }

    private void helpMenuItem_Click(object o, EventArgs e) {
      toggleHelp();
    }

    private void horizontalMenuItem_Click(object o, EventArgs e) {
      horizontal();
    }

    private void verticalMenuItem_Click(object o, EventArgs e) {
      vertical();
    }

    private void opaqueMenuItem_Click(object o, EventArgs e) {
      toggleOpaque();
    }

    private void alwaysOnTopMenuItem_Click(object o, EventArgs e) {
      toggleAlwaysOnTop();
    }

    private void optionsMenuItem_Click(object o, EventArgs e) {
      optionsForm.Set();
      DialogResult result = optionsForm.ShowDialog();
      switch (result) {
      case DialogResult.OK:
        optionsForm.Get();
        BackColor = options.Background;
        if (! isOpaque) {
          Opacity = options.Opacity;
        }
        break;
      case DialogResult.Cancel:
        break;
      }
    }

    private void exitMenuItem_Click(object o, EventArgs e) {
      quit();
    }

    protected override void OnMouseDown(MouseEventArgs e) {
      base.OnMouseDown(e);
      if (e.Button != MouseButtons.Left) {
        return;
      }
      if (isOverHandle(e)) {
        isResizing = true;
        previousX = e.X;
        previousY = e.Y;
        previousLength = length;
      }
      else {
        initialMouseDesktopLocation = PointToScreen(new Point(e.X, e.Y));
        initialDesktopLocation = DesktopLocation;
        isMoving = true;
      }
    }

    protected override void OnMouseMove(MouseEventArgs e) {
      base.OnMouseMove(e);
      if (isMoving) {
        Point currentMouseDesktopLocation = PointToScreen(new Point(e.X, e.Y));
        int deltaX = currentMouseDesktopLocation.X - initialMouseDesktopLocation.X;
        int deltaY = currentMouseDesktopLocation.Y - initialMouseDesktopLocation.Y;
        SetDesktopLocation(initialDesktopLocation.X + deltaX, initialDesktopLocation.Y + deltaY);
      }
      else if (isResizing) {
        int delta;
        if (orientation == Orientation.Horizontal) {
          delta = e.X - previousX;
        }
        else {
          delta = e.Y - previousY;
        }
        length = previousLength + delta;
        resize();
      }
      else {
        if (isOverHandle(e)) {
          if (orientation == Orientation.Horizontal) {
            Cursor = Cursors.SizeWE;
          }
          else {
            Cursor = Cursors.SizeNS;
          }
        }
        else {
          Cursor = Cursors.SizeAll;
        }
      }
    }

    protected override void OnMouseUp(MouseEventArgs e) {
      base.OnMouseUp(e);
      if (e.Button == MouseButtons.Left) {
        isMoving = false;
        isResizing = false;
      }
    }

    protected override void OnKeyPress(KeyPressEventArgs e) {
      base.OnKeyPress(e);
      if (isMoving || isResizing) {
        return;
      }
      switch(e.KeyChar) {
      case 'Q':
      case 'q':
        quit();
        break;
      case 'H':
      case 'h':
        horizontal();
        break;
      case 'V':
      case 'v':
        vertical();
        break;
      case 'O':
      case 'o':
        toggleOpaque();
        break;
      case 'T':
      case 't':
        toggleAlwaysOnTop();
        break;
      default:
        break;
      }
    }

    protected override void OnKeyDown(KeyEventArgs e) {
      base.OnKeyDown(e);
      if (isMoving || isResizing) {
        return;
      }
      int delta = e.Control? 10: 1;
      switch (e.KeyCode) {
      case Keys.Down:
        SetDesktopLocation(DesktopLocation.X, DesktopLocation.Y + delta);
        break;
      case Keys.Up:
        SetDesktopLocation(DesktopLocation.X, DesktopLocation.Y - delta);
        break;
      case Keys.Left:
        SetDesktopLocation(DesktopLocation.X - delta, DesktopLocation.Y);
        break;
      case Keys.Right:
        SetDesktopLocation(DesktopLocation.X + delta, DesktopLocation.Y);
        break;
      case Keys.Oemplus:
        length += delta;
        resize();
        break;
      case Keys.OemMinus:
        length -= delta;
        resize();
        break;
      case Keys.F1:  // FIXME: this should be handled with a menu shortcut
      case Keys.OemQuestion:
        toggleHelp();
        break;
      }
    }

    protected override void OnActivated(EventArgs e) {
      base.OnActivated(e);
      BackColor = options.Background;
    }

    protected override void OnDeactivate(EventArgs e) {
      base.OnDeactivate(e);
      BackColor = DefaultBackColor;
    }

    protected override void OnPaint(PaintEventArgs e) {
      base.OnPaint(e);
      Pen pen = new Pen(options.Foreground);
      Brush brush = new SolidBrush(options.Grip);
      if (orientation == Orientation.Horizontal) {
        e.Graphics.FillRectangle(brush, length - 10, 0, 10, ClientRectangle.Height);
      }
      else {
        e.Graphics.FillRectangle(brush, 0, length - 10, ClientRectangle.Width, 10);
      }
  //    e.Graphics.DrawRectangle(pen, 0, 0, ClientRectangle.Width - 1, ClientRectangle.Height - 1);
      if (orientation == Orientation.Horizontal) {
        for (int i = 0; i < length; i++) {
          if (i % 10 == 0) {
            e.Graphics.DrawLine(pen, i, 0, i, 10);
            e.Graphics.DrawLine(pen, i, breadth - 10, i, breadth);
          }
          else if (i % 2 == 0) {
            e.Graphics.DrawLine(pen, i, 0, i, 5);
            e.Graphics.DrawLine(pen, i, breadth - 5, i, breadth);
          }
          // draw the number
          if (i % 20 == 0) {
            String number = i.ToString();
            SizeF size = e.Graphics.MeasureString(number, privateFont);
            e.Graphics.DrawString(number, privateFont, pen.Brush, i - size.Width / 2, 11);
            e.Graphics.DrawString(number, privateFont, pen.Brush, i - size.Width / 2, breadth - 11 - size.Height);
          }
        }
      }
      else {
        for (int i = 0; i < length; i++) {
          if (i % 10 == 0) {
            e.Graphics.DrawLine(pen, 0, i, 10, i);
            e.Graphics.DrawLine(pen, breadth - 10, i, breadth, i);
          }
          else if (i % 2 == 0) {
            e.Graphics.DrawLine(pen, 0, i, 5, i);
            e.Graphics.DrawLine(pen, breadth - 5, i, breadth, i);
          }
          // draw the number
          if (i % 20 == 0) {
            String number = i.ToString();
            SizeF size = e.Graphics.MeasureString(number, privateFont);
            e.Graphics.DrawString(number, privateFont, pen.Brush, 12, i - size.Height / 2);
            e.Graphics.DrawString(number, privateFont, pen.Brush, breadth - 10 - size.Width, i - size.Height / 2);
          }
        }
      }
      brush.Dispose();
      pen.Dispose();
    }

    [STAThread]
    public static int Main() {
      Application.EnableVisualStyles();
      Application.Run(new Ruler());
      return 0;
    }
  }
}
